@@ -209,6 +209,8 @@ int headerVerifyInfo(int il, int dl, const void * pev, void * iv, int negate)
 	if (end > info->offset)
 	    return i;
 
+	if (info->tag < HEADER_IMAGE)
+	    return i;
 	if (hdrchkType(info->type))
 	    return i;
 	if (hdrchkAlign(info->type, info->offset))
@@ -430,6 +432,9 @@ static int regionSwab(indexEntry entry, int il, int dl,
 	ie.info.count = ntohl(pe->count);
 	ie.info.offset = ntohl(pe->offset);
 
+	if (regionid != 0 && (ie.info.tag >= RPMTAG_HEADERIMAGE &&
+			      ie.info.tag < RPMTAG_HEADERREGIONS))
+	    return -1;
 	if (hdrchkType(ie.info.type))
 	    return -1;
 	if (hdrchkData(ie.info.count))
@@ -1070,7 +1075,6 @@ int headerIsEntry(Header h, rpmTagVal tag)
  * @todo Permit retrieval of regions other than HEADER_IMUTABLE.
  * @param entry		header entry
  * @param td		tag data container
- * @param minMem	string pointers refer to header memory?
  * @param flags		flags to control memory allocation
  * @return		1 on success, otherwise error.
  */
@@ -1187,6 +1191,7 @@ static int copyTdEntry(const indexEntry entry, rpmtd td, headerGetFlags flags)
     }
     td->type = entry->info.type;
     td->count = count;
+    td->size = entry->length;
 
     if (td->data && entry->data != td->data) {
 	td->flags |= RPMTD_ALLOCED;
