@@ -147,16 +147,6 @@ exit:
     return seen;
 }
 
-int parsePGPSig(rpmtd sigtd, const char *type,
-		 pgpDigParams *sig, char **msg)
-{
-    int rc = pgpPrtParams(sigtd->data, sigtd->count, PGPTAG_SIGNATURE, sig);
-
-    if (rc != 0 && type != NULL)
-	rasprintf(msg, _("%s has unverifiable signature"), type);
-    return rc;
-}
-
 /*
  * Argument monster to verify header-only signature/digest if there is
  * one, otherwisereturn RPMRC_NOTFOUND to signal for plain sanity check.
@@ -171,7 +161,7 @@ static rpmRC headerSigVerify(rpmKeyring keyring, rpmVSFlags vsflags,
     pgpDigParams sig = NULL;
     struct rpmtd_s sigtd;
     struct entryInfo_s info, einfo;
-    unsigned int hashalgo = 0;
+    struct sigtInfo_s sinfo;
 
     rpmtdReset(&sigtd);
     memset(&info, 0, sizeof(info));
@@ -246,22 +236,11 @@ static rpmRC headerSigVerify(rpmKeyring keyring, rpmVSFlags vsflags,
     sigtd.data = memcpy(xmalloc(siglen), dataStart + info.offset, siglen);
     sigtd.flags = RPMTD_ALLOCED;
 
-    switch (info.tag) {
-    case RPMTAG_RSAHEADER:
-    case RPMTAG_DSAHEADER:
-	if (parsePGPSig(&sigtd, "header", &sig, buf))
-	    goto exit;
-	hashalgo = pgpDigParamsAlgo(sig, PGPVAL_HASHALGO);
-	break;
-    case RPMTAG_SHA1HEADER:
-	hashalgo = PGPHASHALGO_SHA1;
-	break;
-    default:
-	break;
-    }
+    if (rpmSigInfoParse(&sigtd, "header", &sinfo, &sig, buf))
+	goto exit;
 
-    if (hashalgo) {
-	DIGEST_CTX ctx = rpmDigestInit(hashalgo, RPMDIGEST_NONE);
+    if (sinfo.hashalgo) {
+	DIGEST_CTX ctx = rpmDigestInit(sinfo.hashalgo, RPMDIGEST_NONE);
 	int32_t ildl[2] = { htonl(ril), htonl(rdl) };
 
 	rpmDigestUpdate(ctx, rpm_header_magic, sizeof(rpm_header_magic));
@@ -292,6 +271,12 @@ rpmRC headerVerifyRegion(rpmTagVal regionTag,
     int32_t ril = 0;
     int32_t rdl = 0;
 
+    /* Check that we have at least on tag */
+    if (il < 1) {
+	rasprintf(buf, _("region: no tags"));
+	goto exit;
+    }
+
     memset(entry, 0, sizeof(*entry));
 
     /* Check (and convert) the 1st tag element. */
@@ -533,16 +518,14 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
 			Header * hdrp, unsigned int *keyidp, char **msg)
 {
     pgpDigParams sig = NULL;
-    char buf[8*BUFSIZ];
-    ssize_t count;
     Header sigh = NULL;
     rpmTagVal sigtag;
     struct rpmtd_s sigtd;
+    struct sigtInfo_s sinfo;
     Header h = NULL;
     rpmRC rc = RPMRC_FAIL;	/* assume failure */
     int leadtype = -1;
     headerGetFlags hgeflags = HEADERGET_DEFAULT;
-    DIGEST_CTX ctx = NULL;
 
     if (hdrp) *hdrp = NULL;
 
@@ -567,27 +550,17 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
 	(sigtag == 0 && !(vsflags & (_mask)) && headerIsEntry(sigh, (_tag)))
 
     /*
-     * Figger the most effective available signature.
-     * Prefer signatures over digests, then header-only over header+payload.
+     * Figger the most effective means of verification available, prefer
+     * signatures over digests. Legacy header+payload entries are not used.
      * DSA will be preferred over RSA if both exist because tested first.
-     * Note that NEEDPAYLOAD prevents header+payload signatures and digests.
      */
     sigtag = 0;
     if (_chk(RPMVSF_NODSAHEADER, RPMSIGTAG_DSA)) {
 	sigtag = RPMSIGTAG_DSA;
     } else if (_chk(RPMVSF_NORSAHEADER, RPMSIGTAG_RSA)) {
 	sigtag = RPMSIGTAG_RSA;
-    } else if (_chk(RPMVSF_NODSA|RPMVSF_NEEDPAYLOAD, RPMSIGTAG_GPG)) {
-	sigtag = RPMSIGTAG_GPG;
-	fdInitDigest(fd, PGPHASHALGO_SHA1, 0);
-    } else if (_chk(RPMVSF_NORSA|RPMVSF_NEEDPAYLOAD, RPMSIGTAG_PGP)) {
-	sigtag = RPMSIGTAG_PGP;
-	fdInitDigest(fd, PGPHASHALGO_MD5, 0);
     } else if (_chk(RPMVSF_NOSHA1HEADER, RPMSIGTAG_SHA1)) {
 	sigtag = RPMSIGTAG_SHA1;
-    } else if (_chk(RPMVSF_NOMD5|RPMVSF_NEEDPAYLOAD, RPMSIGTAG_MD5)) {
-	sigtag = RPMSIGTAG_MD5;
-	fdInitDigest(fd, PGPHASHALGO_MD5, 0);
     }
 
     /* Read the metadata, computing digest(s) on the fly. */
@@ -614,55 +587,23 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
 	goto exit;
     }
 
-    switch (sigtag) {
-    case RPMSIGTAG_RSA:
-    case RPMSIGTAG_DSA:
-	if (parsePGPSig(&sigtd, "package", &sig, msg)) {
-	    rc = RPMRC_FAIL;
-	    goto exit;
-	}
-	/* fallthrough */
-    case RPMSIGTAG_SHA1:
-    {	struct rpmtd_s utd;
-	unsigned int hashalgo = (sigtag == RPMSIGTAG_SHA1) ?
-				PGPHASHALGO_SHA1 :
-				pgpDigParamsAlgo(sig, PGPVAL_HASHALGO);
+    if (rpmSigInfoParse(&sigtd, "package", &sinfo, &sig, msg) == RPMRC_OK) {
+	struct rpmtd_s utd;
+	DIGEST_CTX ctx = rpmDigestInit(sinfo.hashalgo, RPMDIGEST_NONE);
 
-	if (!headerGet(h, RPMTAG_HEADERIMMUTABLE, &utd, hgeflags))
-	    break;
-	ctx = rpmDigestInit(hashalgo, RPMDIGEST_NONE);
-	(void) rpmDigestUpdate(ctx, rpm_header_magic, sizeof(rpm_header_magic));
-	(void) rpmDigestUpdate(ctx, utd.data, utd.count);
-	rpmtdFreeData(&utd);
-    }	break;
-    case RPMSIGTAG_GPG:
-    case RPMSIGTAG_PGP5:	/* XXX legacy */
-    case RPMSIGTAG_PGP:
-	if (parsePGPSig(&sigtd, "package", &sig, msg)) {
-	    rc = RPMRC_FAIL;
-	    goto exit;
-	}
-	/* fallthrough */
-    case RPMSIGTAG_MD5:
-	/* Legacy signatures need the compressed payload in the digest too. */
-	while ((count = Fread(buf, sizeof(buf[0]), sizeof(buf), fd)) > 0) {}
-	if (count < 0) {
-	    rasprintf(msg, _("Fread failed: %s"), Fstrerror(fd));
-	    rc = RPMRC_FAIL;
-	    goto exit;
+	if (headerGet(h, RPMTAG_HEADERIMMUTABLE, &utd, hgeflags)) {
+	    rpmDigestUpdate(ctx, rpm_header_magic, sizeof(rpm_header_magic));
+	    rpmDigestUpdate(ctx, utd.data, utd.count);
+	    rpmtdFreeData(&utd);
 	}
+	/** @todo Implement disable/enable/warn/error/anal policy. */
+	rc = rpmVerifySignature(keyring, &sigtd, sig, ctx, msg);
 
-	ctx = rpmDigestBundleDupCtx(fdGetBundle(fd),(sigtag == RPMSIGTAG_MD5) ?
-				    PGPHASHALGO_MD5 :
-				    pgpDigParamsAlgo(sig, PGPVAL_HASHALGO));
-	break;
-    default:
-	break;
+	rpmDigestFinal(ctx, NULL, NULL, 0);
+    } else {
+	rc = RPMRC_FAIL;
     }
 
-    /** @todo Implement disable/enable/warn/error/anal policy. */
-    rc = rpmVerifySignature(keyring, &sigtd, sig, ctx, msg);
-
 exit:
     if (rc != RPMRC_FAIL && h != NULL && hdrp != NULL) {
 	/* Retrofit RPMTAG_SOURCEPACKAGE to srpms for compatibility */
@@ -700,7 +641,6 @@ exit:
 	    *keyidp = getKeyid(sig);
     }
     rpmtdFreeData(&sigtd);
-    rpmDigestFinal(ctx, NULL, NULL, 0);
     h = headerFree(h);
     pgpDigParamsFree(sig);
     sigh = rpmFreeSignature(sigh);
