@@ -17,6 +17,7 @@ typedef struct rpmal_s * rpmal;
 
 /**
  * Initialize available packckages, items, and directory list.
+ * @param pool		shared string pool with base, dir and dependency names
  * @param delta		no. of entries to add on each realloc
  * @param tsflags	transaction control flags
  * @param tscolor	transaction color bits
@@ -24,7 +25,7 @@ typedef struct rpmal_s * rpmal;
  * @return al		new available list
  */
 RPM_GNUC_INTERNAL
-rpmal rpmalCreate(int delta, rpmtransFlags tsflags,
+rpmal rpmalCreate(rpmstrPool pool, int delta, rpmtransFlags tsflags,
 		  rpm_color_t tscolor, rpm_color_t prefcolor);
 
 /**
@@ -52,24 +53,32 @@ RPM_GNUC_INTERNAL
 void rpmalAdd(rpmal al, rpmte p);
 
 /**
- * Lookup best provider for a dependency in the available list
+ * Lookup all obsoleters for a dependency in the available list
+ * @param al		available list
+ * @param ds		dependency set
+ * @return		obsoleting packages for ds, NULL if none
+ */
+RPM_GNUC_INTERNAL
+rpmte * rpmalAllObsoletes(const rpmal al, const rpmds ds);
+
+/**
+ * Lookup all providers for a dependency in the available list
  * @param al		available list
  * @param ds		dependency set
  * @return		best provider for the dependency, NULL if none
  */
 RPM_GNUC_INTERNAL
-rpmte rpmalSatisfiesDepend(const rpmal al, const rpmds ds);
+rpmte * rpmalAllSatisfiesDepend(const rpmal al, const rpmds ds);
 
 /**
- * Get a list of transaction elements that are memebers of a collection in the
- * available list
+ * Lookup best provider for a dependency in the available list
  * @param al		available list
- * @param collname	collection name to search for
- * @return		NULL-terminated list of transaction elements that are
- *			members of the specified collection
+ * @param te		transaction element
+ * @param ds		dependency set
+ * @return		best provider for the dependency, NULL if none
  */
 RPM_GNUC_INTERNAL
-rpmte * rpmalAllInCollection(const rpmal al, const char * collname);
+rpmte rpmalSatisfiesDepend(const rpmal al, const rpmte te, const rpmds ds);
 
 #ifdef __cplusplus
 }
