@@ -91,6 +91,10 @@ enum rpmQVSources_e {
     RPMQV_HDRID,	/*!< ... from header id (immutable header SHA1). */
     RPMQV_TID,		/*!< ... from install transaction id (time stamp). */
     RPMQV_SPECSRPM,	/*!< ... from spec file source (query only). */
+    RPMQV_WHATRECOMMENDS,	/*!< ... from recommends db search. */
+    RPMQV_WHATSUGGESTS,		/*!< ... from suggests db search. */
+    RPMQV_WHATSUPPLEMENTS,	/*!< ... from supplements db search. */
+    RPMQV_WHATENHANCES,		/*!< ... from enhances db search. */
 };
 
 typedef rpmFlags rpmQVSources;
@@ -123,14 +127,15 @@ enum rpmQueryFlags_e {
     QUERY_FOR_STATE	= (1 << 24),	/*!< query:  from --state */
     QUERY_FOR_DOCS	= (1 << 25),	/*!< query:  from --docfiles */
     QUERY_FOR_CONFIG	= (1 << 26),	/*!< query:  from --configfiles */
-    QUERY_FOR_DUMPFILES	= (1 << 27)	/*!< query:  from --dump */
+    QUERY_FOR_DUMPFILES	= (1 << 27),	/*!< query:  from --dump */
+    QUERY_FOR_LICENSE	= (1 << 28)	/*!< query:  from --licensefiles */
 };
 
 typedef rpmFlags rpmQueryFlags;
 
 #define	_QUERY_FOR_BITS	\
    (QUERY_FOR_LIST|QUERY_FOR_STATE|QUERY_FOR_DOCS|QUERY_FOR_CONFIG|\
-    QUERY_FOR_DUMPFILES)
+    QUERY_FOR_LICENSE|QUERY_FOR_DUMPFILES)
 
 /** \ingroup rpmcli
  * Bit(s) from common command line options.
@@ -281,7 +286,8 @@ enum rpmInstallFlags_e {
     INSTALL_FRESHEN	= (1 << 6),	/*!< from --freshen */
     INSTALL_INSTALL	= (1 << 7),	/*!< from --install */
     INSTALL_ERASE	= (1 << 8),	/*!< from --erase */
-    INSTALL_ALLMATCHES	= (1 << 9)	/*!< from --allmatches */
+    INSTALL_ALLMATCHES	= (1 << 9),	/*!< from --allmatches */
+    INSTALL_REINSTALL	= (1 << 10),	/*!< from --reinstall */
 };
 
 typedef rpmFlags rpmInstallFlags;
@@ -342,7 +348,7 @@ struct rpmInstallArguments_s {
 };
 
 /** \ingroup rpmcli
- * Install/upgrade/freshen binary rpm package.
+ * Install/upgrade/freshen/reinstall binary rpm package.
  * @param ts		transaction set
  * @param ia		mode flags and parameters
  * @param fileArgv	array of package file names (NULL terminated)
