@@ -47,7 +47,7 @@
 #undef HTKEYTYPE
 #undef HTDATATYPE
 
-typedef rpmRC (*idxfunc)(dbiCursor dbc, const char *keyp, size_t keylen,
+typedef rpmRC (*idxfunc)(dbiIndex dbi, dbiCursor dbc, const char *keyp, size_t keylen,
 			 dbiIndexItem rec);
 
 static rpmRC tag2index(dbiIndex dbi, rpmTagVal rpmtag,
@@ -75,7 +75,9 @@ static int buildIndexes(rpmdb db)
     /* Don't call us again */
     db->db_buildindex = 0;
 
-    dbSetFSync(db->db_dbenv, 0);
+    dbSetFSync(db, 0);
+
+    dbCtrl(db, RPMDB_CTRL_LOCK_RW);
 
     mi = rpmdbInitIterator(db, RPMDBI_PACKAGES, NULL, 0);
     while ((h = rpmdbNextIterator(mi))) {
@@ -89,7 +91,11 @@ static int buildIndexes(rpmdb db)
 	}
     }
     rpmdbFreeIterator(mi);
-    dbSetFSync(db->db_dbenv, !db->cfg.db_no_fsync);
+
+    dbCtrl(db, DB_CTRL_INDEXSYNC);
+    dbCtrl(db, DB_CTRL_UNLOCK_RW);
+
+    dbSetFSync(db, !db->cfg.db_no_fsync);
     return rc;
 }
 
@@ -154,7 +160,7 @@ static int pkgdbOpen(rpmdb db, int flags, dbiIndex *dbip)
 	if ((!verifyonly && (dbiFlags(dbi) & DBI_CREATED)) || db->cfg.db_no_fsync) {
 	    rpmlog(RPMLOG_DEBUG, "disabling fsync on database\n");
 	    db->cfg.db_no_fsync = 1;
-	    dbSetFSync(db->db_dbenv, 0);
+	    dbSetFSync(db, 0);
 	}
     } else {
 	rpmlog(RPMLOG_ERR, _("cannot open %s index using %s - %s (%d)\n"),
@@ -222,15 +228,12 @@ static rpmRC indexGet(dbiIndex dbi, const char *keyp, size_t keylen,
 	dbiCursor dbc = dbiCursorInit(dbi, DBC_READ);
 
 	if (keyp) {
-	    if (keylen == 0) {
+	    if (keylen == 0)
 		keylen = strlen(keyp);
-		if (keylen == 0)
-		    keylen++; /* XXX "/" fixup */
-	    }
-	    rc = dbcCursorGet(dbc, keyp, keylen, set);
+	    rc = idxdbGet(dbi, dbc, keyp, keylen, set, DBC_NORMAL_SEARCH);
 	} else {
 	    do {
-		rc = dbcCursorGet(dbc, NULL, 0, set);
+		rc = idxdbGet(dbi, dbc, NULL, 0, set, DBC_NORMAL_SEARCH);
 	    } while (rc == RPMRC_OK);
 
 	    /* If we got some results, not found is not an error */
@@ -238,11 +241,29 @@ static rpmRC indexGet(dbiIndex dbi, const char *keyp, size_t keylen,
 		rc = RPMRC_OK;
 	}
 
-	dbiCursorFree(dbc);
+	dbiCursorFree(dbi, dbc);
     }
     return rc;
 }
 
+static rpmRC indexPrefixGet(dbiIndex dbi, const char *pfx, size_t plen,
+			    dbiIndexSet *set)
+{
+    rpmRC rc = RPMRC_FAIL; /* assume failure */
+
+    if (dbi != NULL && pfx) {
+	dbiCursor dbc = dbiCursorInit(dbi, DBC_READ);
+
+	if (plen == 0)
+	    plen = strlen(pfx);
+	rc = idxdbGet(dbi, dbc, pfx, plen, set, DBC_PREFIX_SEARCH);
+
+	dbiCursorFree(dbi, dbc);
+    }
+    return rc;
+}
+
+
 typedef struct miRE_s {
     rpmTagVal		tag;		/*!< header tag */
     rpmMireMode		mode;		/*!< pattern match mode */
@@ -282,6 +303,7 @@ struct rpmdbIndexIterator_s {
     rpmDbiTag		ii_rpmtag;
     dbiCursor		ii_dbc;
     dbiIndexSet		ii_set;
+    unsigned int	*ii_hdrNums;
 };
 
 static rpmdb rpmdbRock;
@@ -311,23 +333,14 @@ int rpmdbCheckTerminate(int terminate)
 	rpmdbMatchIterator mi;
 	rpmdbIndexIterator ii;
 
-	while ((mi = rpmmiRock) != NULL) {
-	    rpmmiRock = mi->mi_next;
-	    mi->mi_next = NULL;
+	while ((mi = rpmmiRock) != NULL)
 	    rpmdbFreeIterator(mi);
-	}
 
-	while ((ii = rpmiiRock) != NULL) {
-	    rpmiiRock = ii->ii_next;
-	    ii->ii_next = NULL;
+	while ((ii = rpmiiRock) != NULL)
 	    rpmdbIndexIteratorFree(ii);
-	}
 
-	while ((db = rpmdbRock) != NULL) {
-	    rpmdbRock = db->db_next;
-	    db->db_next = NULL;
+	while ((db = rpmdbRock) != NULL)
 	    (void) rpmdbClose(db);
-	}
     }
     sigprocmask(SIG_SETMASK, &oldMask, NULL);
     return terminating;
@@ -432,6 +445,12 @@ int rpmdbClose(rpmdb db)
     if (db == NULL)
 	goto exit;
 
+    prev = &rpmdbRock;
+    while ((next = *prev) != NULL && next != db)
+	prev = &next->db_next;
+    if (!next)
+	goto exit;
+
     (void) rpmdbUnlink(db);
 
     if (db->nrefs > 0)
@@ -439,7 +458,7 @@ int rpmdbClose(rpmdb db)
 
     /* Always re-enable fsync on close of rw-database */
     if ((db->db_mode & O_ACCMODE) != O_RDONLY)
-	dbSetFSync(db->db_dbenv, 1);
+	dbSetFSync(db, 1);
 
     if (db->db_pkgs)
 	rc = dbiClose(db->db_pkgs, 0);
@@ -452,9 +471,6 @@ int rpmdbClose(rpmdb db)
     db->db_indexes = _free(db->db_indexes);
     db->db_descr = _free(db->db_descr);
 
-    prev = &rpmdbRock;
-    while ((next = *prev) != NULL && next != db)
-	prev = &next->db_next;
     if (next) {
         *prev = next->db_next;
 	next->db_next = NULL;
@@ -480,7 +496,6 @@ static rpmdb newRpmdb(const char * root, const char * home,
     char * db_home = rpmGetPath((home && *home) ? home : "%{_dbpath}", NULL);
 
     static rpmDbiTag const dbiTags[] = {
-	RPMDBI_PACKAGES,
 	RPMDBI_NAME,
 	RPMDBI_BASENAMES,
 	RPMDBI_GROUP,
@@ -493,6 +508,12 @@ static rpmdb newRpmdb(const char * root, const char * home,
 	RPMDBI_INSTALLTID,
 	RPMDBI_SIGMD5,
 	RPMDBI_SHA1HEADER,
+	RPMDBI_FILETRIGGERNAME,
+	RPMDBI_TRANSFILETRIGGERNAME,
+	RPMDBI_RECOMMENDNAME,
+	RPMDBI_SUGGESTNAME,
+	RPMDBI_SUPPLEMENTNAME,
+	RPMDBI_ENHANCENAME,
     };
 
     if (!(db_home && db_home[0] != '%')) {
@@ -612,7 +633,10 @@ int rpmdbVerify(const char * prefix)
 	int xx;
 	rc = rpmdbOpenAll(db);
 
-	rc = dbiForeach(db->db_indexes, db->db_ndbi, dbiVerify, 0);
+	
+	if (db->db_pkgs)
+	    rc += dbiVerify(db->db_pkgs, 0);
+	rc += dbiForeach(db->db_indexes, db->db_ndbi, dbiVerify, 0);
 
 	xx = rpmdbClose(db);
 	if (xx && rc == 0) rc = xx;
@@ -621,7 +645,7 @@ int rpmdbVerify(const char * prefix)
     return rc;
 }
 
-static Header rpmdbGetHeaderAt(rpmdb db, unsigned int offset)
+Header rpmdbGetHeaderAt(rpmdb db, unsigned int offset)
 {   
     rpmdbMatchIterator mi = rpmdbInitIterator(db, RPMDBI_PACKAGES,
 					      &offset, sizeof(offset));
@@ -1019,8 +1043,11 @@ static int miFreeHeader(rpmdbMatchIterator mi, dbiIndex dbi)
 	    sigset_t signalMask;
 
 	    blockSignals(&signalMask);
+	    dbCtrl(mi->mi_db, DB_CTRL_LOCK_RW);
 	    rc = pkgdbPut(dbi, mi->mi_dbc, mi->mi_prevoffset,
-			  hdrBlob, hdrLen, NULL);
+			  hdrBlob, hdrLen);
+	    dbCtrl(mi->mi_db, DB_CTRL_INDEXSYNC);
+	    dbCtrl(mi->mi_db, DB_CTRL_UNLOCK_RW);
 	    unblockSignals(&signalMask);
 
 	    if (rc) {
@@ -1052,13 +1079,14 @@ rpmdbMatchIterator rpmdbFreeIterator(rpmdbMatchIterator mi)
     if (next) {
 	*prev = next->mi_next;
 	next->mi_next = NULL;
-    }
+    } else
+	return NULL;
 
     pkgdbOpen(mi->mi_db, 0, &dbi);
 
     miFreeHeader(mi, dbi);
 
-    mi->mi_dbc = dbiCursorFree(mi->mi_dbc);
+    mi->mi_dbc = dbiCursorFree(dbi, mi->mi_dbc);
 
     if (mi->mi_re != NULL)
     for (i = 0; i < mi->mi_nre; i++) {
@@ -1097,6 +1125,24 @@ int rpmdbGetIteratorCount(rpmdbMatchIterator mi)
     return (mi && mi->mi_set ?  mi->mi_set->count : 0);
 }
 
+int rpmdbGetIteratorIndex(rpmdbMatchIterator mi)
+{
+    return (mi ? mi->mi_setx : 0);
+}
+
+void rpmdbSetIteratorIndex(rpmdbMatchIterator mi, unsigned int ix)
+{
+    if (mi)
+	mi->mi_setx = ix;
+}
+
+unsigned int rpmdbGetIteratorOffsetFor(rpmdbMatchIterator mi, unsigned int ix)
+{
+    if (mi && mi->mi_set && ix < mi->mi_set->count)
+	return mi->mi_set->recs[ix].hdrNum;
+    return 0;
+}
+
 /**
  * Return pattern match.
  * @param mire		match iterator regex
@@ -1511,7 +1557,9 @@ top:
 	    mi->mi_offset = dbiIndexRecordOffset(mi->mi_set, mi->mi_setx);
 	    mi->mi_filenum = dbiIndexRecordFileNumber(mi->mi_set, mi->mi_setx);
 	} else {
-	    rc = pkgdbGet(dbi, mi->mi_dbc, 0, &uh, &uhlen, &mi->mi_offset);
+	    rc = pkgdbGet(dbi, mi->mi_dbc, 0, &uh, &uhlen);
+	    if (rc == 0)
+		mi->mi_offset = pkgdbKey(dbi, mi->mi_dbc);
 
 	    /* Terminate on error or end of keys */
 	    if (rc || (mi->mi_setx && mi->mi_offset == 0))
@@ -1526,7 +1574,7 @@ top:
 
     /* Retrieve next header blob for index iterator. */
     if (uh == NULL) {
-	rc = pkgdbGet(dbi, mi->mi_dbc, mi->mi_offset, &uh, &uhlen, NULL);
+	rc = pkgdbGet(dbi, mi->mi_dbc, mi->mi_offset, &uh, &uhlen);
 	if (rc)
 	    return NULL;
     }
@@ -1579,6 +1627,13 @@ void rpmdbSortIterator(rpmdbMatchIterator mi)
     }
 }
 
+void rpmdbUniqIterator(rpmdbMatchIterator mi)
+{
+    if (mi && mi->mi_set) {
+	dbiIndexSetUniq(mi->mi_set, mi->mi_sorted);
+    }
+}
+
 int rpmdbExtendIterator(rpmdbMatchIterator mi,
 			const void * keyp, size_t keylen)
 {
@@ -1604,22 +1659,31 @@ int rpmdbExtendIterator(rpmdbMatchIterator mi,
     return rc;
 }
 
-int rpmdbPruneIterator(rpmdbMatchIterator mi, removedHash hdrNums)
+int rpmdbFilterIterator(rpmdbMatchIterator mi, packageHash hdrNums, int neg)
 {
-    if (mi == NULL || hdrNums == NULL || removedHashNumKeys(hdrNums) <= 0)
+    if (mi == NULL || hdrNums == NULL)
 	return 1;
 
     if (!mi->mi_set)
-        return 0;
+	return 0;
+
+    if (packageHashNumKeys(hdrNums) == 0) {
+	if (!neg)
+	    mi->mi_set->count = 0;
+	return 0;
+    }
 
     unsigned int from;
     unsigned int to = 0;
     unsigned int num = mi->mi_set->count;
+    int cond;
 
     assert(mi->mi_set->count > 0);
 
     for (from = 0; from < num; from++) {
-	if (removedHashHasEntry(hdrNums, mi->mi_set->recs[from].hdrNum)) {
+	cond = !packageHashHasEntry(hdrNums, mi->mi_set->recs[from].hdrNum);
+	cond = neg ? !cond : cond;
+	if (cond) {
 	    mi->mi_set->count--;
 	    continue;
 	}
@@ -1630,6 +1694,15 @@ int rpmdbPruneIterator(rpmdbMatchIterator mi, removedHash hdrNums)
     return 0;
 }
 
+int rpmdbPruneIterator(rpmdbMatchIterator mi, packageHash hdrNums)
+{
+    if (packageHashNumKeys(hdrNums) <= 0)
+	return 1;
+
+    return rpmdbFilterIterator(mi, hdrNums, 1);
+}
+
+
 int rpmdbAppendIterator(rpmdbMatchIterator mi,
 			const unsigned int * hdrNums, unsigned int nHdrNums)
 {
@@ -1650,8 +1723,13 @@ rpmdbMatchIterator rpmdbNewIterator(rpmdb db, rpmDbiTagVal dbitag)
 {
     rpmdbMatchIterator mi = NULL;
 
-    if (indexOpen(db, dbitag, 0, NULL))
-	return NULL;
+    if (dbitag == RPMDBI_PACKAGES) {
+	if (pkgdbOpen(db, 0, NULL))
+	    return NULL;
+    } else {
+	if (indexOpen(db, dbitag, 0, NULL))
+	    return NULL;
+    }
 
     mi = xcalloc(1, sizeof(*mi));
     mi->mi_set = NULL;
@@ -1764,6 +1842,40 @@ rpmdbMatchIterator rpmdbInitIterator(rpmdb db, rpmDbiTagVal rpmtag,
     return mi;
 }
 
+rpmdbMatchIterator rpmdbInitPrefixIterator(rpmdb db, rpmDbiTagVal rpmtag,
+					    const void * pfx, size_t plen)
+{
+    rpmdbMatchIterator mi = NULL;
+    dbiIndexSet set = NULL;
+    dbiIndex dbi = NULL;
+    rpmDbiTagVal dbtag = rpmtag;
+
+    if (!pfx)
+	return NULL;
+
+    if (db != NULL && rpmtag != RPMDBI_PACKAGES) {
+	(void) rpmdbCheckSignals();
+
+
+	if (indexOpen(db, dbtag, 0, &dbi) == 0) {
+	    int rc = 0;
+
+	    rc = indexPrefixGet(dbi, pfx, plen, &set);
+
+	    if (rc)	{
+		set = dbiIndexSetFree(set);
+	    } else {
+		mi = rpmdbNewIterator(db, dbtag);
+		mi->mi_set = set;
+		rpmdbSortIterator(mi);
+	    }
+	}
+
+    }
+
+    return mi;
+}
+
 /*
  * Convert current tag data to db key
  * @param tagdata	Tag data container
@@ -1804,8 +1916,6 @@ static const void * td2key(rpmtd tagdata, unsigned int *keylen)
 	str = rpmtdGetString(tagdata);
 	if (str) {
 	    size = strlen(str);
-	    if (size == 0)
-		size++; /* fixup for empty strings */
 	    data = str;
 	}
 	break;
@@ -1862,9 +1972,9 @@ int rpmdbIndexIteratorNext(rpmdbIndexIterator ii, const void ** key, size_t * ke
     /* free old data */
     ii->ii_set = dbiIndexSetFree(ii->ii_set);
 
-    rc = dbcCursorGet(ii->ii_dbc, NULL, 0, &ii->ii_set);
+    rc = idxdbGet(ii->ii_dbi, ii->ii_dbc, NULL, 0, &ii->ii_set, DBC_NORMAL_SEARCH);
 
-    *key = dbiCursorKey(ii->ii_dbc, &iikeylen);
+    *key = idxdbKey(ii->ii_dbi, ii->ii_dbc, &iikeylen);
     *keylen = iikeylen;
 
     return (rc == RPMRC_OK) ? 0 : -1;
@@ -1930,6 +2040,24 @@ unsigned int rpmdbIndexIteratorPkgOffset(rpmdbIndexIterator ii, unsigned int nr)
     return dbiIndexRecordOffset(ii->ii_set, nr);
 }
 
+unsigned int *rpmdbIndexIteratorPkgOffsets(rpmdbIndexIterator ii)
+{
+    int i;
+
+    if (!ii || !ii->ii_set)
+	return NULL;
+
+    if (ii->ii_hdrNums)
+	ii->ii_hdrNums = _free(ii->ii_hdrNums);
+
+    ii->ii_hdrNums = xmalloc(sizeof(*ii->ii_hdrNums) * ii->ii_set->count);
+    for (i = 0; i < ii->ii_set->count; i++) {
+	ii->ii_hdrNums[i] = ii->ii_set->recs[i].hdrNum;
+    }
+
+    return ii->ii_hdrNums;
+}
+
 unsigned int rpmdbIndexIteratorTagNum(rpmdbIndexIterator ii, unsigned int nr)
 {
     if (!ii || !ii->ii_set)
@@ -1952,13 +2080,17 @@ rpmdbIndexIterator rpmdbIndexIteratorFree(rpmdbIndexIterator ii)
     if (next) {
         *prev = next->ii_next;
         next->ii_next = NULL;
-    }
+    } else
+	return NULL;
 
-    ii->ii_dbc = dbiCursorFree(ii->ii_dbc);
+    ii->ii_dbc = dbiCursorFree(ii->ii_dbi, ii->ii_dbc);
     ii->ii_dbi = NULL;
     rpmdbClose(ii->ii_db);
     ii->ii_set = dbiIndexSetFree(ii->ii_set);
 
+    if (ii->ii_hdrNums)
+	ii->ii_hdrNums = _free(ii->ii_hdrNums);
+
     ii = _free(ii);
     return NULL;
 }
@@ -1979,12 +2111,13 @@ static void logAddRemove(const char *dbiname, int removing, rpmtd tagdata)
 
 static rpmRC indexDel(dbiIndex dbi, rpmTagVal rpmtag, unsigned int hdrNum, Header h)
 {
-    return tag2index(dbi, rpmtag, hdrNum, h, dbcCursorDel);
+    return tag2index(dbi, rpmtag, hdrNum, h, idxdbDel);
 }
 
 int rpmdbRemove(rpmdb db, unsigned int hdrNum)
 {
     dbiIndex dbi = NULL;
+    dbiCursor dbc = NULL;
     Header h;
     sigset_t signalMask;
     int ret = 0;
@@ -2008,9 +2141,12 @@ int rpmdbRemove(rpmdb db, unsigned int hdrNum)
 	return 1;
 
     (void) blockSignals(&signalMask);
+    dbCtrl(db, DB_CTRL_LOCK_RW);
 
     /* Remove header from primary index */
-    ret = pkgdbDel(dbi, NULL, hdrNum);
+    dbc = dbiCursorInit(dbi, DBC_WRITE);
+    ret = pkgdbDel(dbi, dbc, hdrNum);
+    dbiCursorFree(dbi, dbc);
 
     /* Remove associated data from secondary indexes */
     if (ret == 0) {
@@ -2024,6 +2160,8 @@ int rpmdbRemove(rpmdb db, unsigned int hdrNum)
 	}
     }
 
+    dbCtrl(db, DB_CTRL_INDEXSYNC);
+    dbCtrl(db, DB_CTRL_UNLOCK_RW);
     (void) unblockSignals(&signalMask);
 
     headerFree(h);
@@ -2032,22 +2170,109 @@ int rpmdbRemove(rpmdb db, unsigned int hdrNum)
     return 0;
 }
 
+struct updateRichDepData {
+    ARGV_t argv;
+    int nargv;
+    int neg;
+    int level;
+    int *nargv_level;
+};
+
+static rpmRC updateRichDepCB(void *cbdata, rpmrichParseType type,
+		const char *n, int nl, const char *e, int el, rpmsenseFlags sense,
+		rpmrichOp op, char **emsg) {
+    struct updateRichDepData *data = cbdata;
+    if (type == RPMRICH_PARSE_ENTER) {
+	data->level++;
+	data->nargv_level = xrealloc(data->nargv_level, data->level * (sizeof(int)));
+	data->nargv_level[data->level - 1] = data->nargv;
+    }
+    if (type == RPMRICH_PARSE_LEAVE) {
+	data->level--;
+    }
+    if (type == RPMRICH_PARSE_SIMPLE && nl && !(nl > 7 && !strncmp(n, "rpmlib(", 7))) {
+	char *name = xmalloc(nl + 2);
+	*name = data->neg ? '!' : ' ';
+	strncpy(name + 1, n, nl);
+	name[1 + nl] = 0;
+	argvAdd(&data->argv, name);
+	data->nargv++;
+	_free(name);
+    }
+    if (type == RPMRICH_PARSE_OP && op == RPMRICHOP_IF) {
+	/* save nargv in case of ELSE */
+	data->nargv_level[data->level - 1] = data->nargv;
+	data->neg ^= 1;
+    }
+    if (type == RPMRICH_PARSE_OP && op == RPMRICHOP_ELSE) {
+	int i, nargv = data->nargv;
+	/* copy and invert condition block */
+	for (i = data->nargv_level[data->level - 1]; i < nargv; i++) {
+	    char *name = data->argv[i];
+	    *name ^= ' ' ^ '!';
+	    argvAdd(&data->argv, name);
+	    *name ^= ' ' ^ '!';
+	    data->nargv++;
+	}
+	data->neg ^= 1;
+    }
+    if (type == RPMRICH_PARSE_LEAVE && op == RPMRICHOP_IF) {
+	data->neg ^= 1;
+    }
+    return RPMRC_OK;
+}
+
+static rpmRC updateRichDep(dbiIndex dbi, dbiCursor dbc, const char *str,
+                           struct dbiIndexItem_s *rec,
+                           idxfunc idxupdate)
+{
+    int n, i, rc = 0;
+    struct updateRichDepData data;
+
+    data.argv = argvNew();
+    data.neg = 0;
+    data.nargv = 0;
+    data.level = 0;
+    data.nargv_level = xcalloc(1, sizeof(int));
+    if (rpmrichParse(&str, NULL, updateRichDepCB, &data) == RPMRC_OK) {
+	n = argvCount(data.argv);
+	if (n) {
+	    argvSort(data.argv, NULL);
+	    for (i = 0; i < n; i++) {
+		char *name = data.argv[i];
+		if (i && !strcmp(data.argv[i - 1], name))
+		    continue;       /* ignore dups */
+		if (*name == ' ')
+		    name++;
+		rc += idxupdate(dbi, dbc, name, strlen(name), rec);
+	    }
+	}
+    }
+    _free(data.nargv_level);
+    argvFree(data.argv);
+    return rc;
+}
+
 static rpmRC tag2index(dbiIndex dbi, rpmTagVal rpmtag,
 		       unsigned int hdrNum, Header h,
 		       idxfunc idxupdate)
 {
     int i, rc = 0;
-    struct rpmtd_s tagdata, reqflags;
+    struct rpmtd_s tagdata, reqflags, trig_index;
     dbiCursor dbc = NULL;
 
     switch (rpmtag) {
     case RPMTAG_REQUIRENAME:
 	headerGet(h, RPMTAG_REQUIREFLAGS, &reqflags, HEADERGET_MINMEM);
-	/* fallthrough */
-    default:
-	headerGet(h, rpmtag, &tagdata, HEADERGET_MINMEM);
+	break;
+    case RPMTAG_FILETRIGGERNAME:
+	headerGet(h, RPMTAG_FILETRIGGERINDEX, &trig_index, HEADERGET_MINMEM);
+	break;
+    case RPMTAG_TRANSFILETRIGGERNAME:
+	headerGet(h, RPMTAG_TRANSFILETRIGGERINDEX, &trig_index, HEADERGET_MINMEM);
 	break;
     }
+    headerGet(h, rpmtag, &tagdata, HEADERGET_MINMEM);
 
     if (rpmtdCount(&tagdata) == 0) {
 	if (rpmtag != RPMTAG_GROUP)
@@ -2066,8 +2291,22 @@ static rpmRC tag2index(dbiIndex dbi, rpmTagVal rpmtag,
 	const void * key = NULL;
 	unsigned int keylen = 0;
 	int j;
-	/* Include the tagNum in all indices (only files use though) */
-	struct dbiIndexItem_s rec = { .hdrNum = hdrNum, .tagNum = i };
+	struct dbiIndexItem_s rec;
+
+	switch (rpmtag) {
+	/* Include trigger index in db index for triggers */
+	case RPMTAG_FILETRIGGERNAME:
+	case RPMTAG_TRANSFILETRIGGERNAME:
+	    rec.hdrNum = hdrNum;
+	    rec.tagNum = *rpmtdNextUint32(&trig_index);
+	    break;
+
+	/* Include the tagNum in the others indices (only files use though) */
+	default:
+	    rec.hdrNum = hdrNum;
+	    rec.tagNum = i;
+	    break;
+	}
 
 	switch (rpmtag) {
 	case RPMTAG_REQUIRENAME: {
@@ -2097,24 +2336,43 @@ static rpmRC tag2index(dbiIndex dbi, rpmTagVal rpmtag,
 	if ((key = td2key(&tagdata, &keylen)) == NULL)
 	    continue;
 
-	rc += idxupdate(dbc, key, keylen, &rec);
+	rc += idxupdate(dbi, dbc, key, keylen, &rec);
+
+	if (*(char *)key == '(') {
+	    switch (rpmtag) {
+	    case RPMTAG_REQUIRENAME:
+	    case RPMTAG_CONFLICTNAME:
+	    case RPMTAG_SUGGESTNAME:
+	    case RPMTAG_SUPPLEMENTNAME:
+	    case RPMTAG_RECOMMENDNAME:
+	    case RPMTAG_ENHANCENAME:
+		if (rpmtdType(&tagdata) == RPM_STRING_ARRAY_TYPE) {
+		    rc += updateRichDep(dbi, dbc, rpmtdGetString(&tagdata),
+			&rec, idxupdate);
+		}
+	    default:
+		break;
+	    }
+	}
     }
 
-    dbiCursorFree(dbc);
+    dbiCursorFree(dbi, dbc);
 
 exit:
     rpmtdFreeData(&tagdata);
     return (rc == 0) ? RPMRC_OK : RPMRC_FAIL;
 }
+
 static rpmRC indexPut(dbiIndex dbi, rpmTagVal rpmtag, unsigned int hdrNum, Header h)
 {
-    return tag2index(dbi, rpmtag, hdrNum, h, dbcCursorPut);
+    return tag2index(dbi, rpmtag, hdrNum, h, idxdbPut);
 }
 
 int rpmdbAdd(rpmdb db, Header h)
 {
     sigset_t signalMask;
     dbiIndex dbi = NULL;
+    dbiCursor dbc = NULL;
     unsigned int hdrNum = 0;
     unsigned int hdrLen = 0;
     unsigned char *hdrBlob = NULL;
@@ -2134,9 +2392,14 @@ int rpmdbAdd(rpmdb db, Header h)
 	goto exit;
 	
     (void) blockSignals(&signalMask);
+    dbCtrl(db, DB_CTRL_LOCK_RW);
 
     /* Add header to primary index */
-    ret = pkgdbPut(dbi, NULL, 0, hdrBlob, hdrLen, &hdrNum);
+    dbc = dbiCursorInit(dbi, DBC_WRITE);
+    ret = pkgdbNew(dbi, dbc, &hdrNum);
+    if (ret == 0)
+	ret = pkgdbPut(dbi, dbc, hdrNum, hdrBlob, hdrLen);
+    dbiCursorFree(dbi, dbc);
 
     /* Add associated data to secondary indexes */
     if (ret == 0) {	
@@ -2150,6 +2413,8 @@ int rpmdbAdd(rpmdb db, Header h)
 	}
     }
 
+    dbCtrl(db, DB_CTRL_INDEXSYNC);
+    dbCtrl(db, DB_CTRL_UNLOCK_RW);
     (void) unblockSignals(&signalMask);
 
     /* If everything ok, mark header as installed now */
@@ -2224,10 +2489,9 @@ static int rpmdbRemoveDatabase(const char * prefix, const char * dbpath)
 
 static int renameTag(const char * prefix,
 		     const char * olddbpath, const char *newdbpath,
-		     rpmTagVal dbtag)
+		     const char * base)
 {
     int xx, rc = 0;
-    const char *base = rpmTagGetName(dbtag);
     char *src = rpmGetPath(prefix, "/", olddbpath, "/", base, NULL);
     char *dest = rpmGetPath(prefix, "/", newdbpath, "/", base, NULL);
     struct stat st;
@@ -2265,10 +2529,14 @@ static int rpmdbMoveDatabase(const char * prefix,
     rpmdb db = newRpmdb(prefix, newdbpath, O_RDONLY, 0644, RPMDB_FLAG_REBUILD);
 
     blockSignals(&sigMask);
-    rc = renameTag(prefix, olddbpath, newdbpath, RPMDBI_PACKAGES);
+    rc = renameTag(prefix, olddbpath, newdbpath, rpmTagGetName(RPMDBI_PACKAGES));
     for (int i = 0; i < db->db_ndbi; i++) {
-	rc += renameTag(prefix, olddbpath, newdbpath, db->db_tags[i]);
+	rc += renameTag(prefix, olddbpath, newdbpath, rpmTagGetName(db->db_tags[i]));
     }
+#ifdef ENABLE_NDB
+    rc += renameTag(prefix, olddbpath, newdbpath, "Packages.db");
+    rc += renameTag(prefix, olddbpath, newdbpath, "Index.db");
+#endif
 
     cleanDbenv(prefix, olddbpath);
     cleanDbenv(prefix, newdbpath);
@@ -2407,3 +2675,27 @@ exit:
 
     return rc;
 }
+
+int rpmdbCtrl(rpmdb db, rpmdbCtrlOp ctrl)
+{
+    dbCtrlOp dbctrl = 0;
+    switch (ctrl) {
+    case RPMDB_CTRL_LOCK_RO:
+	dbctrl = DB_CTRL_LOCK_RO;
+	break;
+    case RPMDB_CTRL_UNLOCK_RO:
+	dbctrl = DB_CTRL_UNLOCK_RO;
+	break;
+    case RPMDB_CTRL_LOCK_RW:
+	dbctrl = DB_CTRL_LOCK_RW;
+	break;
+    case RPMDB_CTRL_UNLOCK_RW:
+	dbctrl = DB_CTRL_UNLOCK_RW;
+	break;
+    case RPMDB_CTRL_INDEXSYNC:
+	dbctrl = DB_CTRL_INDEXSYNC;
+	break;
+    }
+    return dbctrl ? dbCtrl(db, dbctrl) : 1;
+}
+
