@@ -50,6 +50,9 @@ struct rpmfi_s {
     char * fn;			/*!< File name buffer. */
     char * ofn;			/*!< Original file name buffer. */
 
+    int intervalStart;		/*!< Start of iterating interval. */
+    int intervalEnd;		/*!< End of iterating interval. */
+
     rpmfiles files;		/*!< File info set */
     rpmcpio_t archive;		/*!< Archive with payload */
     unsigned char * found;	/*!< Bit field of files found in the archive */
@@ -111,7 +114,9 @@ struct rpmfiles_s {
     struct fingerPrint_s * fps;	/*!< File fingerprint(s). */
 
     int digestalgo;		/*!< File digest algorithm */
+    int signaturelength;	/*!< File signature length */
     unsigned char * digests;	/*!< File digests in binary. */
+    unsigned char * signatures; /*!< File signatures in binary. */
 
     struct nlinkHash_s * nlinks;/*!< Files connected by hardlinks */
     rpm_off_t * replacedSizes;	/*!< (TR_ADDED) */
@@ -475,6 +480,12 @@ int rpmfiFindOFN(rpmfi fi, const char * fn)
     return ix;
 }
 
+static int cmpPfx(rpmfiles files, int ix, const char *pfx)
+{
+    int plen = strlen(pfx);
+    return strncmp(pfx, rpmfilesDN(files, rpmfilesDI(files, ix)), plen);
+}
+
 rpmfileAttrs rpmfilesFFlags(rpmfiles fi, int ix)
 {
     rpmfileAttrs FFlags = 0;
@@ -560,6 +571,19 @@ char * rpmfiFDigestHex(rpmfi fi, int *algo)
     return fdigest;
 }
 
+const unsigned char * rpmfilesFSignature(rpmfiles fi, int ix, size_t *len)
+{
+	const unsigned char *signature = NULL;
+
+	if (fi != NULL && ix >= 0 && ix < rpmfilesFC(fi)) {
+	     if (fi->signatures != NULL)
+		signature = fi->signatures + (fi->signaturelength * ix);
+	     if (len)
+		*len = fi->signaturelength;
+	}
+	return signature;
+}
+
 const char * rpmfilesFLink(rpmfiles fi, int ix)
 {
     const char * flink = NULL;
@@ -817,6 +841,16 @@ static int iterBack(rpmfi fi)
     return fi->i - 1;
 }
 
+static int iterInterval(rpmfi fi)
+{
+    if (fi->i == -1)
+	return fi->intervalStart;
+    else if (fi->i + 1 < fi->intervalEnd)
+	return fi->i + 1;
+    else
+	return RPMERR_ITER_END;
+}
+
 int rpmfiNext(rpmfi fi)
 {
     int next = -1;
@@ -1146,6 +1180,7 @@ rpmfiles rpmfilesFree(rpmfiles fi)
 	fi->flinks = _free(fi->flinks);
 	fi->flangs = _free(fi->flangs);
 	fi->digests = _free(fi->digests);
+	fi->signatures = _free(fi->signatures);
 	fi->fcaps = _free(fi->fcaps);
 
 	fi->cdict = _free(fi->cdict);
@@ -1360,7 +1395,7 @@ static int rpmfilesPopulate(rpmfiles fi, Header h, rpmfiFlags flags)
     headerGetFlags scareFlags = (flags & RPMFI_KEEPHEADER) ? 
 				HEADERGET_MINMEM : HEADERGET_ALLOC;
     headerGetFlags defFlags = HEADERGET_ALLOC;
-    struct rpmtd_s fdigests, digalgo, td;
+    struct rpmtd_s fdigests, fsignatures, digalgo, td;
     unsigned char * t;
 
     /* XXX TODO: all these should be sanity checked, ugh... */
@@ -1411,6 +1446,8 @@ static int rpmfilesPopulate(rpmfiles fi, Header h, rpmfiFlags flags)
 	}
     }
 
+    fi->signaturelength = headerGetNumber(h, RPMTAG_FILESIGNATURELENGTH);
+
     fi->digests = NULL;
     /* grab hex digests from header and store in binary format */
     if (!(flags & RPMFI_NOFILEDIGESTS) &&
@@ -1431,6 +1468,25 @@ static int rpmfilesPopulate(rpmfiles fi, Header h, rpmfiFlags flags)
 	rpmtdFreeData(&fdigests);
     }
 
+    fi->signatures = NULL;
+    /* grab hex signatures from header and store in binary format */
+    if (! (flags & RPMFI_NOFILESIGNATURES) &&
+	headerGet(h, RPMTAG_FILESIGNATURES, &fsignatures, HEADERGET_MINMEM)) {
+	const char *fsignature;
+	fi->signatures = t = xmalloc(rpmtdCount(&fsignatures) * fi->signaturelength);
+
+	while ((fsignature = rpmtdNextString(&fsignatures))) {
+	    if (*fsignature == '\0') {
+		memset(t, 0, fi->signaturelength);
+		t += fi->signaturelength;
+		continue;
+	    }
+	    for (int j = 0; j < fi->signaturelength; j++, t++, fsignature += 2)
+		*t = (rnibble(fsignature[0]) << 4) | rnibble(fsignature[1]);
+	}
+	rpmtdFreeData(&fsignatures);
+    }
+
     /* XXX TR_REMOVED doesn;t need fmtimes, frdevs, finodes */
     if (!(flags & RPMFI_NOFILEMTIMES))
 	_hgfi(h, RPMTAG_FILEMTIMES, &td, scareFlags, fi->fmtimes);
@@ -1512,6 +1568,7 @@ int (*nextfuncs[])(rpmfi fi) = {
     iterReadArchiveNext,
     iterReadArchiveNextContentFirst,
     iterReadArchiveNextOmitHardlinks,
+    iterInterval,
 };
 
 
@@ -1527,7 +1584,9 @@ static rpmfi initIter(rpmfiles files, int itype, int link)
 	fi->i = -1;
 	if (itype == RPMFI_ITER_BACK) {
 	    fi->i = rpmfilesFC(fi->files);
-	} else if (itype >=RPMFI_ITER_READ_ARCHIVE) {
+	} else if (itype >=RPMFI_ITER_READ_ARCHIVE
+	    && itype <= RPMFI_ITER_READ_ARCHIVE_OMIT_HARDLINKS) {
+
 	    fi->found = xcalloc(1, (rpmfiFC(fi)>>3) + 1);
 	}
 	rpmfiLink(fi);
@@ -1541,6 +1600,50 @@ rpmfi rpmfilesIter(rpmfiles files, int itype)
     return initIter(files, itype, 1);
 }
 
+rpmfi rpmfilesFindPrefix(rpmfiles fi, const char *pfx)
+{
+    int l, u, c, comparison;
+    rpmfi iterator = NULL;
+
+    if (!fi || !pfx)
+	return NULL;
+
+    l = 0;
+    u = rpmfilesFC(fi);
+    while (l < u) {
+	c = (l + u) / 2;
+
+	comparison = cmpPfx(fi, c, pfx);
+
+	if (comparison < 0)
+	    u = c;
+	else if (comparison > 0)
+	    l = c + 1;
+	else {
+	    if (cmpPfx(fi, l, pfx))
+		l = c;
+	    while (l > 0 && !cmpPfx(fi, l - 1, pfx))
+		l--;
+	    if ( u >= rpmfilesFC(fi) || cmpPfx(fi, u, pfx))
+		u = c;
+	    while (++u < rpmfilesFC(fi)) {
+		if (cmpPfx(fi, u, pfx))
+		    break;
+	    }
+	    break;
+	}
+
+    }
+
+    if (l < u) {
+	iterator = initIter(fi, RPMFI_ITER_INTERVAL, 1);
+	iterator->intervalStart = l;
+	iterator->intervalEnd = u;
+    }
+
+    return iterator;
+}
+
 rpmfi rpmfiNewPool(rpmstrPool pool, Header h, rpmTagVal tagN, rpmfiFlags flags)
 {
     rpmfiles files = rpmfilesNew(pool, h, tagN, flags);
@@ -1662,6 +1765,11 @@ const unsigned char * rpmfiFDigest(rpmfi fi, int *algo, size_t *len)
     return rpmfilesFDigest(fi->files, fi ? fi->i : -1, algo, len);
 }
 
+const unsigned char * rpmfiFSignature(rpmfi fi, size_t *len)
+{
+    return rpmfilesFSignature(fi->files, fi ? fi->i : -1, len);
+}
+
 uint32_t rpmfiFDepends(rpmfi fi, const uint32_t ** fddictp)
 {
     return rpmfilesFDepends(fi->files,  fi ? fi->i : -1, fddictp);
