@@ -17,6 +17,8 @@
 
 #include "debug.h"
 
+#define MANIFEST_RECURSIONS 1000 /* Max. number of allowed manifest recursions */
+
 RPM_GNUC_INTERNAL
 rpmgiFlags giFlags = RPMGI_NONE;
 
@@ -31,6 +33,10 @@ struct rpmgi_s {
 
     ARGV_t argv;
     int argc;
+
+    int curLvl;			/*!< Current recursion level */
+    int	recLvls[MANIFEST_RECURSIONS]; /*!< Reversed end index for given level */
+
 };
 
 /**
@@ -50,7 +56,7 @@ static FD_t rpmgiOpen(const char * path, const char * fmode)
 	if (fd != NULL) (void) Fclose(fd);
 	fd = NULL;
     }
-    fn = _free(fn);
+    free(fn);
 
     return fd;
 }
@@ -123,11 +129,24 @@ static Header rpmgiLoadReadHeader(rpmgi gi)
     if (gi->argv != NULL && gi->argv[gi->i] != NULL)
     do {
 	char * fn = gi->argv[gi->i];
-	int rc = rpmgiReadHeader(gi, fn, &h);
+	int rc;
+
+	while (gi->recLvls[gi->curLvl] > gi->argc - gi->i)
+	    gi->curLvl--;
+
+	rc = rpmgiReadHeader(gi, fn, &h);
 
 	if (h != NULL || (gi->flags & RPMGI_NOMANIFEST) || rc == 0)
 	    break;
 
+	if (gi->curLvl == MANIFEST_RECURSIONS - 1) {
+	    rpmlog(RPMLOG_ERR,
+		   _("Max level of manifest recursion exceeded: %s\n"), fn);
+	    break;
+	}
+	gi->curLvl++;
+	gi->recLvls[gi->curLvl] = gi->argc - gi->i;
+
 	/* Not a header, so try for a manifest. */
 	gi->argv[gi->i] = NULL;		/* Mark the insertion point */
 	if (rpmgiLoadManifest(gi, fn) != RPMRC_OK) {
@@ -150,46 +169,38 @@ static Header rpmgiLoadReadHeader(rpmgi gi)
  */
 static void rpmgiGlobArgv(rpmgi gi, ARGV_const_t argv)
 {
-    const char * arg;
-    int ac = 0;
-    int xx;
+    if (argv == NULL) return;
 
     /* XXX Expand globs only if requested */
     if ((gi->flags & RPMGI_NOGLOB)) {
-	if (argv != NULL) {
-	    while (argv[ac] != NULL)
-		ac++;
-	    xx = argvAppend(&gi->argv, argv);
+	argvAppend(&gi->argv, argv);
+    } else {
+	const char * arg;
+	while ((arg = *argv++) != NULL) {
+	    char * t = rpmEscapeSpaces(arg);
+	    char ** av = NULL;
+
+	    if (rpmGlob(t, NULL, &av) == 0) {
+		argvAppend(&gi->argv, av);
+		argvFree(av);
+	    }
+	    free(t);
 	}
-	gi->argc = ac;
-	return;
     }
+    gi->argc = argvCount(gi->argv);
 
-    if (argv != NULL)
-    while ((arg = *argv++) != NULL) {
-	char * t = rpmEscapeSpaces(arg);
-	char ** av = NULL;
-
-	xx = rpmGlob(t, &ac, &av);
-	xx = argvAppend(&gi->argv, av);
-	gi->argc += ac;
-	av = argvFree(av);
-	t = _free(t);
-	ac = 0;
-    }
     return;
 }
 
 rpmgi rpmgiFree(rpmgi gi)
 {
-    if (gi == NULL)
-	return NULL;
+    if (gi != NULL) {
+	rpmtsFree(gi->ts);
+	argvFree(gi->argv);
 
-    gi->ts = rpmtsFree(gi->ts);
-    gi->argv = argvFree(gi->argv);
-
-    memset(gi, 0, sizeof(*gi));		/* XXX trash and burn */
-    gi = _free(gi);
+	memset(gi, 0, sizeof(*gi)); /* XXX trash and burn */
+	free(gi);
+    }
     return NULL;
 }
 
@@ -208,6 +219,9 @@ rpmgi rpmgiNew(rpmts ts, rpmgiFlags flags, ARGV_const_t argv)
     gi->argc = 0;
     rpmgiGlobArgv(gi, argv);
 
+    gi->curLvl = 0;
+    gi->recLvls[gi->curLvl] = 1;
+
     return gi;
 }
 
