@@ -8,13 +8,16 @@
 #include "lib/fprint.h"
 #include "lib/rpmlock.h"
 #include "lib/rpmdb_internal.h"
+#include "lib/rpmscript.h"
+#include "lib/rpmtriggers.h"
 
 typedef struct diskspaceInfo_s * rpmDiskSpaceInfo;
 
 /* Transaction set elements information */
 typedef struct tsMembers_s {
     rpmstrPool pool;		/*!< Global string pool */
-    removedHash removedPackages;	/*!< Set of packages being removed. */
+    packageHash removedPackages;	/*!< Set of packages being removed. */
+    packageHash installedPackages;	/*!< Set of installed packages */
     rpmal addedPackages;	/*!< Set of packages being installed. */
 
     rpmds rpmlib;		/*!< rpmlib() dependency set. */
@@ -24,6 +27,17 @@ typedef struct tsMembers_s {
     int delta;			/*!< Delta for reallocation. */
 } * tsMembers;
 
+typedef struct tsTrigger_s {
+    unsigned int hdrNum;
+    int index;
+} tsTrigger;
+
+typedef struct tsTriggers_s {
+    tsTrigger *trigger;
+    int count;
+    int alloced;
+} tsTriggers;
+
 /** \ingroup rpmts
  * The set of packages to be installed/removed atomically.
  */
@@ -47,10 +61,9 @@ struct rpmts_s {
 
     tsMembers members;		/*!< Transaction set member info (order etc) */
 
-    struct selabel_handle * selabelHandle;	/*!< Handle to selabel */
-
     char * rootDir;		/*!< Path to top of install tree. */
     char * lockPath;		/*!< Transaction lock path */
+    rpmlock lock;		/*!< Transaction lock file */
     FD_t scriptFd;		/*!< Scriptlet stdout/stderr. */
     rpm_tid_t tid;		/*!< Transaction id. */
 
@@ -69,6 +82,8 @@ struct rpmts_s {
     rpmPlugins plugins;		/*!< Transaction plugins */
 
     int nrefs;			/*!< Reference count. */
+
+    rpmtriggers trigs2run;   /*!< Transaction file triggers */
 };
 
 #ifdef __cplusplus
@@ -86,6 +101,11 @@ rpmstrPool rpmtsPool(rpmts ts);
 RPM_GNUC_INTERNAL
 tsMembers rpmtsMembers(rpmts ts);
 
+/* Return rpmdb iterator with removals optionally pruned out */
+RPM_GNUC_INTERNAL
+rpmdbMatchIterator rpmtsPrunedIterator(rpmts ts, rpmDbiTagVal tag,
+					      const char * key, int prune);
+
 RPM_GNUC_INTERNAL
 rpmal rpmtsCreateAl(rpmts ts, rpmElementTypes types);
 
@@ -94,29 +114,11 @@ RPM_GNUC_INTERNAL
 int rpmtsSolve(rpmts ts, rpmds key);
 
 RPM_GNUC_INTERNAL
-rpmlock rpmtsAcquireLock(rpmts ts);
+rpmRC rpmtsSetupTransactionPlugins(rpmts ts);
 
-/** \ingroup rpmts
- * Get the selabel handle from the transaction set
- * @param ts		transaction set
- * @return		rpm selabel handle, or NULL if it hasn't been initialized yet
- */
-struct selabel_handle * rpmtsSELabelHandle(rpmts ts);
-
-/** \ingroup rpmts
- * Initialize selabel
- * @param ts		transaction set
- * @param open_status   if the func should open selinux status or just check it
- * @return		RPMRC_OK on success, RPMRC_FAIL otherwise
- */
-rpmRC rpmtsSELabelInit(rpmts ts, int open_status);
-
-/** \ingroup rpmts
- * Clean up selabel
- * @param ts		transaction set
- * @param close_status  whether we should close selinux status
- */
-void rpmtsSELabelFini(rpmts ts, int close_status);
+RPM_GNUC_INTERNAL
+rpmRC runScript(rpmts ts, rpmte te, ARGV_const_t prefixes,
+		       rpmScript script, int arg1, int arg2);
 
 #ifdef __cplusplus
 }
