@@ -308,21 +308,25 @@ static SECKEYPublicKey *pgpNewPublicKey(KeyType type)
 static int pgpSetSigMpiDSA(pgpDigAlg pgpsig, int num, const uint8_t *p)
 {
     SECItem *sig = pgpsig->data;
-    int lbits = DSA1_Q_BITS;
+    unsigned int qbits = DSA1_Q_BITS;
+    unsigned int subprlen = DSA1_SUBPRIME_LEN;
+    unsigned int siglen = DSA1_SIGNATURE_LEN;
     int rc = 1; /* assume failure */
 
     switch (num) {
     case 0:
-	sig = pgpsig->data = SECITEM_AllocItem(NULL, NULL, DSA1_SIGNATURE_LEN);
+	sig = pgpsig->data = SECITEM_AllocItem(NULL, NULL, siglen);
 	if (sig) {
-	    memset(sig->data, 0, DSA1_SIGNATURE_LEN);
-	    rc = pgpMpiSet(lbits, sig->data, p);
+	    memset(sig->data, 0, siglen);
+	    rc = pgpMpiSet(qbits, sig->data, p);
 	}
 	break;
     case 1:
-	if (sig && pgpMpiSet(lbits, sig->data+DSA1_SUBPRIME_LEN, p) == 0) {
+	if (sig && pgpMpiSet(qbits, sig->data+subprlen, p) == 0) {
 	    SECItem *signew = SECITEM_AllocItem(NULL, NULL, 0);
-	    if (signew && DSAU_EncodeDerSig(signew, sig) == SECSuccess) {
+	    if (signew == NULL)
+		break;
+	    if (DSAU_EncodeDerSigWithLen(signew, sig, siglen) == SECSuccess) {
 		SECITEM_FreeItem(sig, PR_TRUE);
 		pgpsig->data = signew;
 		rc = 0;
