@@ -19,6 +19,7 @@
 #include <errno.h>
 #include <popt.h>
 #include <ctype.h>
+#include <pthread.h>
 
 #include <rpm/rpmfileutil.h>
 #include <rpm/rpmurl.h>
@@ -31,6 +32,7 @@
 #include "debug.h"
 
 static const char *rpm_config_dir = NULL;
+static pthread_once_t configDirSet = PTHREAD_ONCE_INIT;
 
 static int is_prelinked(int fdno)
 {
@@ -337,7 +339,8 @@ int rpmFileIsCompressed(const char * file, rpmCompressedMagic * compressed)
 
     rc = 0;
 
-    if ((magic[0] == 'B') && (magic[1] == 'Z')) {
+    if ((magic[0] == 'B') && (magic[1] == 'Z') &&
+        (magic[2] == 'h')) {
 	*compressed = COMPRESSED_BZIP2;
     } else if ((magic[0] == 'P') && (magic[1] == 'K') &&
 	 (((magic[2] == 3) && (magic[3] == 4)) ||
@@ -610,11 +613,14 @@ int rpmMkdirs(const char *root, const char *pathstr)
     return rc;
 }
 
+static void setConfigDir(void)
+{
+    char *rpmenv = getenv("RPM_CONFIGDIR");
+    rpm_config_dir = rpmenv ? xstrdup(rpmenv) : RPMCONFIGDIR;
+}
+
 const char *rpmConfigDir(void)
 {
-    if (rpm_config_dir == NULL) {
-	char *rpmenv = getenv("RPM_CONFIGDIR");
-	rpm_config_dir = rpmenv ? xstrdup(rpmenv) : RPMCONFIGDIR;
-    }
+    pthread_once(&configDirSet, setConfigDir);
     return rpm_config_dir;
 }
